#!/usr/bin/env Rscript

library(GA)
library(plyr)
library(parallelMap)
library(lightgbm)


# analyze parameters from command line
Args <- commandArgs()
parallelStartSocket(as.integer(Args[8]))

# define a function for preliminary parameter tuning in step 1 
tuneModel_one_grid <- function(xData,yData,params,grid){
  max_auc = 0.5
  best = 0
  for(i in 1:nrow(grid)){
      # analyze parameters from function
      params[[names(grid)]] = grid[i,]
      lgb_weight <- (yData * params$weight + 1) / sum(yData * params$weight + 1)
      lgb_train <- lgb.Dataset(
          data = data.matrix(xData),
          label = yData, 
          free_raw_data = TRUE,
          weight = lgb_weight
      )   
      # cross validation
      lgb_tr_mod <- lgb.cv(
          params,
          data = lgb_train,
          nrounds = 300,
          stratified = TRUE,
          nfold = 10,
          num_threads = as.integer(Args[8]),
          early_stopping_rounds = 10
      )   
      # obtain an AUC value for this iteration
      auc = unlist(lgb_tr_mod$record_evals$valid$auc$eval)[length(unlist(lgb_tr_mod$record_evals$valid$auc$eval))]
      # compare this AUC value with the current best AUC value
      if(auc > max_auc){
          max_auc = auc
          best = grid[i,]
     }    
  }
  # return optimized parameters
  params[[names(grid)]] = best
  return(params)
}       

# define a fitness function for GA-based parameter tuning in step 2
tuneModel_GA <- function(weight,learningRate, numLeaves,maxDepth, minDataInLeaf, maxBin, featureFraction, minSumHessian, lambdaL1, lambdaL2, dropRate, maxDrop){
  lgb_weight <- (data.y *as.integer(weight) + 1) / sum(data.y * as.integer(weight) + 1)
  
  lgb_train <- lgb.Dataset(
    data = data.matrix(data.x), 
    label = data.y, 
    free_raw_data = TRUE,
    weight = lgb_weight
  )
  
  # receive parameters from function
  params <- list(
    objective = 'binary',
    metric = 'auc',
    learning_rate = learningRate,
    num_leaves = as.integer(numLeaves),
    max_depth = as.integer(maxDepth),
    min_data_in_leaf = as.integer(minDataInLeaf),
    max_bin = as.integer(maxBin),                               # must be integer
    feature_fraction = featureFraction,
    min_sum_hessian = minSumHessian,
    lambda_l1 = lambdaL1,
    lambda_l2 = lambdaL2,
    drop_rate = dropRate,
    max_drop = as.integer(maxDrop)
  )
  
  # cross validation
  lgb_tr_mod <- lgb.cv(
    params,
    data = lgb_train,
    nrounds = 300,
    stratified = TRUE,
    nfold = 10,
    num_threads = as.integer(Args[8]),
    early_stopping_rounds = 10
  )

  # return AUC value for this iteration
  return(unlist(lgb_tr_mod$record_evals$valid$auc$eval)[length(unlist(lgb_tr_mod$record_evals$valid$auc$eval))])
}


if(length(Args)!=9){
  cat("Please input four parameters: input file (in csv format), output file, numbers of CPU cores and parameter tuning strategy (1: step 1 only; 2: step 1 + step 2).\n")
}else{

  featureFile = Args[6]
  outputFile = Args[7]
  status = Args[9]
  LightGBMArgs = vector()
  data = read.csv(featureFile, sep=",", header = TRUE)

  data.x <- data[,c(-1)]
  data.y <- data$Class

  data.y=mapvalues(data.y,from=c(-1),to=c(0))

  tune_params <- list(
    objective = 'binary',
    metric = 'auc'
  )

  # begin to tune parameters in step 1:
  cat("Begin to tune parameters in step 1: \n")
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(weight = 1:10))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(learning_rate = 2^(-(10:1))))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(num_leaves = seq(50, 800, 50)))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(max_depth = seq(5,10,1)))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(min_data_in_leaf = 2 ^ (1:6)))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(max_bin = 2 ^ (5:10)))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(feature_fraction = seq(.5, 1, .02)))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(min_sum_hessian = seq(0, .02, .001)))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(lambda_l1 = seq(0, .01, .002)))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(lambda_l2 = seq(0, .01, .002)))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(drop_rate = seq(0, 1, .1)))
  tune_params = tuneModel_one_grid(data.x,data.y,tune_params,expand.grid(max_drop = seq(1, 30, 2)))

  argRow <- c("LightGBM", "step1", tune_params$weight, tune_params$learning_rate, tune_params$num_leaves,tune_params$max_depth,tune_params$min_data_in_leaf,tune_params$max_bin,tune_params$feature_fraction,tune_params$min_sum_hessian,tune_params$lambda_l1,tune_params$lambda_l2,tune_params$drop_rate,tune_params$max_drop)
  LightGBMArgs <- rbind(LightGBMArgs, argRow)
  cat("Parameter tuning in step 1 finished!\n")
  
  # begin to tune parameters in step 2: 
  alpha <- 1.8
  beta <- 0.2
  upper_boundary_value1 <- 1/alpha
  lower_boundary_value1 <- 1/beta
  if(status == '2'){
    cat("Begin to tune parameters in step 2: \n")
    GA <- ga(type = "real-valued", 
             fitness =  function(x) + tuneModel_GA(x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10], x[11], x[12]),
             min = beta*c(tune_params$weight,tune_params$learning_rate, tune_params$num_leaves, tune_params$max_depth, max(tune_params$min_data_in_leaf,lower_boundary_value1), tune_params$max_bin, tune_params$feature_fraction, tune_params$min_sum_hessian, tune_params$lambda_l1, tune_params$lambda_l2, tune_params$drop_rate, max(tune_params$max_drop,lower_boundary_value1)),
             max = alpha*c(tune_params$weight,tune_params$learning_rate, tune_params$num_leaves, tune_params$max_depth, tune_params$min_data_in_leaf, tune_params$max_bin, min(tune_params$feature_fraction,upper_boundary_value1), tune_params$min_sum_hessian, tune_params$lambda_l1, tune_params$lambda_l2, min(tune_params$drop_rate,upper_boundary_value1), tune_params$max_drop),
             popSize = 50, maxiter = 100, run = 100,names=c("weight","learningRate", "numLeaves","max_depth", "minDataInLeaf", "maxBin", "featureFraction", "minSumHessian", "lambdaL1", "lambdaL2", "dropRate", "maxDrop"),
             optim = TRUE)
    summary(GA)
    column = summary(GA)[-1]$solution
    column = lapply(column,function(x) sprintf("%.3f",x))
    argRow <- c("LightGBM", "step2", column)
    LightGBMArgs <- rbind(LightGBMArgs, argRow)
    cat("Parameter tuning in step 2 finished!\n")
  }
  # write results into output file
  cat("Write results into output file...\n")
  colnames(LightGBMArgs) <- c("Model","Step","weight","learningRate", "numLeaves", "maxDepth","minDataInLeaf", "maxBin", "featureFraction", "minSumHessian", "lambdaL1", "lambdaL2", "dropRate", "maxDrop")
  write.table(LightGBMArgs, file=outputFile, sep=",", row.names = F, col.names = T)

}
